// Basic "File Strip" derived from Cstrip
// Switchable UTF/8-bit
// 06/1/23

#include <string.h>
#include <stdio.h>
#include "filestrip.h"
#include "utfutil.h"


// #define RXDEBUG

#ifdef	RXDEBUG
 #define DPRINTF(s) printf s; fflush(stdout);
#else
 #define DPRINTF(s)
#endif

// for temporary quick-checks...
// Always on:
#define XDPRINTF(s) printf s; fflush(stdout);
// Always off:
#define ZDPRINTF(s)


struct FileStripBuf {
	FileStripBuf *next;
	long base_pos, length, top_pos;	// characters (not bytes)
	char *bufend;
	char buff[FSTRIP_BUFLEN+3];	// allow for UTF & term
	FileStripBuf() : next(NULL) {
			ZDPRINTF((" created buffer at %p\n", buff);)
			// will be filled separately
	}
	int fill(FileStrip *strip) {
		if (!strip->f) return 0;
		int n = fread(buff, 1, FSTRIP_BUFLEN, strip->f);
		DPRINTF(("fread of %d chars\n", n);)
		bufend = buff+n;
		if (strip->useutf && n == FSTRIP_BUFLEN) {	// otherwise assume EOF
			// is it UTF? -- max of 3 bytes
			if ((*(bufend-2) & 0xe0) == 0xe0	// incomplete 3-byte group
				&& fread(bufend++, 1, 1, strip->f) == 1) n++;	// one more needed
			else if ((*(bufend-1) & 0xe0) == 0xe0	// start of 3-byte group
				&& fread(bufend++, 1, 2, strip->f) == 2) n+=2, bufend++;	// get 2 more
			else if ((*(bufend-1) & 0xe0) == 0xc0	// 2-byte group
				&& fread(bufend++, 1, 1, strip->f) == 1) n++;
		}
		DPRINTF(("EOF status is %s\n", feof(strip->f)? "true" : "false");)
		*bufend = '\0';
		base_pos = strip->topbuff;
		length = strip->useutf? utflen(buff) : n;
		top_pos = strip->topbuff += length;
		if (length) {	// actually read something
			if (strip->listend) strip->listend->next = this;
			if (!strip->buflist) strip->buflist = this;
			strip->listend = this;
		}
		return length;
	}
	~FileStripBuf() {
	}
};



FileStrip::FileStrip(FILE * file, bool utf)
	: f(NULL), buflist(NULL), listend(NULL),
	freelist(NULL), segProc(NULL), useutf(utf) 
{
	setFile(file);
}

FileStrip::~FileStrip()
{
	if (f && f != stdin) fclose(f);
	while (FileStripBuf *bp = buflist) {
		DPRINTF(("deleting %p from buflist (next %p)\n", bp, bp->next);)
		buflist = bp->next;
		delete bp;
	}
	while (FileStripBuf *bp = freelist) {
		DPRINTF(("deleting %p from freelist (next %p)\n", bp, bp->next);)
		freelist = bp->next;
		delete bp;
	}
}

int FileStrip::get()
{
	DPRINTF(("get with cptr=%p cpos=%d\n", cptr, cpos);)
	if (!cptr || cptr >= curbuf->bufend) 	// this is OK even for UTF8
		stepBuf(cptr);
	if (!cptr) return CSTRIP_EOF;
	ZDPRINTF(("get  %d  %x '%c'\n", cpos, (*cptr & 0xFF), *cptr);)
	cpos++;
	return useutf? utf2istep(cptr) : (*cptr++ & 0xFF);
}

int FileStrip::look()
{
	if (!cptr || cptr >= curbuf->bufend) 	// this is OK even for UTF8
		stepBuf(cptr);
	if (!cptr) return CSTRIP_EOF;
	ZDPRINTF(("curr  %d  %x '%c'\n", cpos, (*cptr & 0xFF), *cptr);)
	return useutf? utf2i(cptr) : (*cptr & 0xFF);
}


long FileStrip::pos()
{
	return f? cpos : CSTRIP_ERROR;
}


bool FileStrip::pos(long p)
{
	DPRINTF(("trying to set FileStrip pos %d file %p topbuff %d\n",
		 p, f, topbuff);)
	if (!f || p > topbuff)	// note: p *may* be at topbuff! (will force new read)
		return false;
	if (p < lowbuff) {
		DPRINTF(("lowbuff is %d buflist %p... moving\n", lowbuff, buflist);)
		if (!buflist || p < buflist->base_pos) return false;
		curbuf = buflist;
		while (curbuf && p >= curbuf->top_pos) curbuf = curbuf->next;
	}
	if (curbuf) {	// I'm really being paranoid...
		cpos = p;
		if (useutf) {
			cptr = curbuf->buff;
			utfstep(cptr, p - curbuf->base_pos);
		} else cptr = curbuf->buff+(p - curbuf->base_pos);
		DPRINTF(("setting pos %d in prev buffer %p\n", cpos, curbuf);)
	}
	if (cpos < lowbound) lowbound = cpos;	// insurance
	return true;
}



bool FileStrip::lowBound(long p)
{
	DPRINTF(("setting lowbound = %d  b.base_pos = %d  b.top_pos = %d\n",
		 p, buflist? buflist->base_pos:-11, buflist? buflist->top_pos:-11);)
	if (f && (p >= lowbuff || buflist && p >= buflist->base_pos)
		 && p < topbuff) lowbound = p;
	else return false;
	if (lowbound > buflist->top_pos	// buflist is base of buffer chain
		&& keeplow < buflist->top_pos-1) {	// always need to keep one back for scan
		DPRINTF(("Dump from %d(keeplow+1) to %d(top_pos) (lowbound=%d current base=%d)\n",
			keeplow+1, buflist->top_pos, lowbound, buflist->base_pos);)
		if (procDiscard) processSeg(keeplow+1, buflist->top_pos);
		keepAfter(buflist->top_pos-1);
	}
	return true;
}

bool FileStrip::keepAfter(long p) {
	DPRINTF(("setting keeplow = %d (lowbound = %d)\n", p, lowbound);)
	DPRINTF(("base_pos=%d topbuff=%d)\n",
		 buflist? buflist->base_pos : -11, topbuff);)
	if (f && buflist && p>= (buflist->base_pos-1) && p <= topbuff)
		keeplow = p;
	else return false;
	trimBufs();
	return true;
}

bool FileStrip::setFile(FILE * file)
{
	if (f && f != stdin) fclose(f);
	f = NULL;
	while (FileStripBuf *bp = buflist) {
		buflist = bp->next;
		delete bp;
	}
	listend = NULL;
	lowbound = cpos = topbuff = 0;	 // nothing there yet
	DPRINTF(("setting file %p\n", file);)
	if (!file) return true;	// just a close is OK
	f = file;
	curbuf = NULL;
	cptr =  fillBuf();	// prime the pump
	return cpos;	// as bool
}

char * FileStrip::stepBuf(char *&cp)
{
	DPRINTF(("stepBuf with cp %p (curbuf=%p bufend=%p buflist=%p)\n",
		cp, curbuf, curbuf? curbuf->bufend : NULL, buflist);)
	if (!cp || cp == curbuf->bufend) cp = fillBuf();
	else for (FileStripBuf * fbp=buflist; fbp; fbp=fbp->next) { 
		DPRINTF(("looking at buffer %p end %p next %p (buff=%p)\n",
				fbp, fbp->bufend, fbp->next, fbp->next->buff);)
		if (cp == fbp->bufend) {
			cp = fbp->next->buff;
			break; 
		}
	} 
	DPRINTF(("cp stepped to %p\n", cp);)
	return cp; 
}

char * FileStrip::fillBuf()
{
	if (!f) return NULL;
	if (curbuf && curbuf->next) {
		DPRINTF(("no fill needed -- curr was %p, now is %p\n",
		  curbuf, curbuf->next);)
		curbuf = curbuf->next;
	}
	else {
		FileStripBuf * b;
		if (freelist) {
			b = freelist;
			freelist = b->next;
			b->next = NULL;
			DPRINTF(("reusing free buffer %p\n", b);)
			// pointers adjusted (or buffer deleted) by fill below
		}
		else {
			b = new FileStripBuf();
			DPRINTF(("created new buffer %p\n", b);)
		}
		if (feof(f) || !b->fill(this)) {	// EOF
			DPRINTF(("got EOF trying to fill %p\n", b);)
			delete b;
			return NULL;
		} 
		if (curbuf) curbuf->next = b;
		curbuf = b;
		DPRINTF(("new buffer -- now curr is %p\n", curbuf);)
	}
	lowbuff = curbuf->base_pos;
	topbuff = curbuf->top_pos;
	DPRINTF(("fillbuf returns with lowbuff=%d topbuff=%d\n",
		lowbuff, topbuff);)
	return curbuf->buff;
}

void FileStrip::trimBufs()
{
	FileStripBuf *bp;
	while ((bp = buflist) && bp->top_pos <= keeplow && bp->top_pos <= cpos) {
		DPRINTF(("dropping buf %p %d-%d\n",bp, bp->base_pos, bp->top_pos);)
		buflist = bp->next;
		bp->next = freelist;
		freelist = bp;
		if (curbuf == bp) {	// in case we trim past curbuf
			curbuf = NULL;
			cptr = NULL;
		}
		if (!buflist) listend = NULL;
	}
}

//////////////////////////////////////////////////////////////////

bool FileStrip::processSeg(long start, long end) {
	if (!segProc) return false;
	DPRINTF(("processing Seg from %d to %d\n", start, end);)
	if (start > end || !buflist || start < buflist->base_pos || end > topbuff)
		return false;	// can't do that...
	if (start == end) return true;	// for convenience
	FileStripBuf *bp = buflist;
	while (bp && start >= bp->top_pos) bp = bp->next;	// locate start buf
	if (!bp) return false;	// off the end...
	DPRINTF(("found start buffer %p base_pos %d top_pos %d\n",
		 bp, bp->base_pos, bp->top_pos);)
	char *segbase = bp->buff;
	if (useutf)
		utfstep(segbase, start - bp->base_pos);	// step pointer to actual start char
	else segbase += start - bp->base_pos;
	DPRINTF(("found  segbase %p\n", segbase);)
	int seglen;
	while (end > bp->top_pos) {
		seglen = bp->bufend - segbase;
		DPRINTF(("subseg %p %d\n", segbase, seglen);)
		if (seglen) segProc(segbase, seglen);
		bp = bp->next;
		segbase = bp->buff;	// all the buffer for the rest
	}
	char *endp = bp->buff;
	if (useutf)
		utfstep(endp, end - bp->base_pos);	// step pointer to actual end point
	else endp += end - bp->base_pos;
	seglen = endp - segbase;
	DPRINTF(("final subseg %p %d\n", segbase, seglen);)
	if (seglen) segProc(segbase, seglen);
	return true;
}


char * FileStrip::text(long start, long end) {
	DPRINTF(("copying text from %d to %d\n", start, end);)
	DPRINTF((" -- state: lowbuff = %d topbuff %d buflist %p base %d\n",
		lowbuff, topbuff, buflist, buflist->base_pos);)
	if (start > end || !buflist || start < buflist->base_pos || end > topbuff)
		return NULL;	// can't do that...
	FileStripBuf *bp = buflist;
	while (bp && start >= bp->top_pos) bp = bp->next;	// locate start buf
	FileStripBuf *startbp = bp;
	long lentotal = 0;
	char *startp = bp->buff;
	if (useutf)
		utfstep(startp, start - bp->base_pos);	// step pointer to actual start char
	else startp += start - bp->base_pos;
	char *segbase = startp;	// initially
	while (end > bp->top_pos) {
		lentotal += bp->bufend - segbase;
		bp = bp->next;
		segbase = bp->buff;	// all the buffer for the rest
	}
	char *endp = bp->buff;
	if (useutf)
		utfstep(endp, end - bp->base_pos);	// step pointer to actual end point
	else endp += end - bp->base_pos;
	lentotal += endp - segbase;	// final byte count of text
	char *str = new char[lentotal+1];
	DPRINTF(("allocated %d(+1) character block\n", lentotal);)
	char *tptr = str;
	bp = startbp;	// back to first in chain
	segbase = startp;
	long seglen;
	while (end > bp->top_pos) {
		seglen = bp->bufend - segbase;
		memcpy(tptr, segbase, seglen);
		tptr += seglen;
		bp = bp->next;
		segbase = bp->buff;	// all the buffer for the rest
	}
	seglen = endp - segbase;
	memcpy(tptr, segbase, seglen);
	tptr += seglen;
	*tptr = '\0';
	return str;
}


